import openai
from openai.types.chat import ChatCompletionMessageParam

try:
    from agenteasy.models import AIModelBase
except:
    from .models import AIModelBase
from asyncio import Semaphore
import inspect
import tiktoken
from functools import wraps
from .logutil import logger
from typing import TypeVar
from .tools import batch_make_tools, call_tool
import json

T = TypeVar("T")


def deco_storage(wrap_func=None, *, ex=129600):
    """用redis保存生成数据，同时控制内容过期时间"""

    def outer_wrapper(func):

        @wraps(func)
        async def cowrapper(*args, **kwargs):
            # 检测是否存在redis，如果存在一律直接返回结果，不存在再生成
            sig = inspect.signature(func)
            bind_args = sig.bind(*args, **kwargs)
            bind_args.apply_defaults()

            _msgs = bind_args.kwargs.get("messages")
            if not bind_args.args[0].storage_backend:
                result = await func(*bind_args.args, **bind_args.kwargs)
                return result

            # 请求的内容是否已经存在
            content = bind_args.args[0].storage_backend.get(str(_msgs))
            _use_cache = bind_args.kwargs.get("use_backend_cache")
            if content and _use_cache:
                logger.debug(f"content exist, use redis cache")
                return content.decode("utf-8")

            result = await func(*bind_args.args, **bind_args.kwargs)
            bind_args.args[0].storage_backend.set(str(_msgs), result, ex=ex)
            return result

        @wraps(func)
        def wrapper(*args, **kwargs):
            # 检测是否存在redis，如果存在一律直接返回结果，不存在再生成
            sig = inspect.signature(func)
            bind_args = sig.bind(*args, **kwargs)
            bind_args.apply_defaults()

            _msgs = bind_args.kwargs.get("messages")
            if not bind_args.args[0].storage_backend:
                result = func(*bind_args.args, **bind_args.kwargs)
                return result

            # 请求的内容是否已经存在
            content = bind_args.args[0].storage_backend.get(str(_msgs))
            _use_cache = bind_args.kwargs.get("use_backend_cache")
            if content and _use_cache:
                logger.debug(f"content exist, use redis cache")
                return content.decode("utf-8")

            result = func(*bind_args.args, **bind_args.kwargs)
            bind_args.args[0].storage_backend.set(str(_msgs), result, ex=ex)
            return result

        if inspect.iscoroutinefunction(func):
            return cowrapper
        return wrapper

    if wrap_func is None:
        return outer_wrapper
    return outer_wrapper(wrap_func)


def deco_token_limit(wrap_func):
    @wraps(wrap_func)
    async def cowrapper(*args, **kwargs):
        msg_num = count_tokens_from_messages(
            args[0]._token_counter, kwargs.get("messages")
        )
        logger.debug(f"message token estimate: {msg_num}")
        if msg_num > args[0].ai_model.tokens_limit:
            raise ValueError(
                f"{msg_num} tokens exceed the limit of {args[0].ai_model.tokens_limit}"
            )
        result = await wrap_func(*args, **kwargs)
        return result

    @wraps(wrap_func)
    def wrapper(*args, **kwargs):
        msg_num = count_tokens_from_messages(
            args[0]._token_counter, kwargs.get("messages")
        )
        logger.debug(f"message token estimate: {msg_num}")
        if msg_num > args[0].ai_model.tokens_limit:
            raise ValueError(
                f"{msg_num} tokens exceed the limit of {args[0].ai_model.tokens_limit}"
            )
        result = wrap_func(*args, **kwargs)
        return result

    if inspect.iscoroutinefunction(wrap_func):
        return cowrapper
    return wrapper


def count_tokens_from_messages(token_counter, messages):
    num_tokens = 0
    for message in messages:
        num_tokens += (
            4  # every message follows <im_start>{role/name}\n{content}<im_end>\n
        )
        for key, value in message.items():
            num_tokens += len(token_counter.encode(str(value)))
            if key == "name":  # if there's a name, the role is omitted
                num_tokens += -1  # role is always required and always 1 token
    num_tokens += 2  # every reply is primed with <im_start>assistant
    return num_tokens


class AIAgent:
    def __init__(
        self,
        ai_model: AIModelBase,
        model_async: bool = False,
        semaphore_num: int = 3,
        **model_kwargs,
    ) -> None:
        self.ai_model = ai_model
        self.client = None
        self.aclient = None
        self._semaphore: Semaphore | None = None
        self._store_backend = None
        self._token_counter = tiktoken.get_encoding(self.ai_model.token_counter)
        self.semaphore_num = semaphore_num
        if model_async:
            self.create_aclient(**model_kwargs)
        else:
            self.create_client(**model_kwargs)

    def __del__(self):
        if self._store_backend:
            self._store_backend.close()

    @property
    def storage_backend(self):
        return self._store_backend

    def create_client(self, **kwargs):
        logger.debug(f"client {self.ai_model.model} created")
        self.client = openai.OpenAI(
            api_key=self.ai_model.api_key, base_url=self.ai_model.base_url, **kwargs
        )

    def create_aclient(self, **kwargs):
        logger.debug(f"async client {self.ai_model.model} created")
        self.aclient = openai.AsyncClient(
            api_key=self.ai_model.api_key, base_url=self.ai_model.base_url, **kwargs
        )
        self._semaphore = Semaphore(self.semaphore_num)

    @deco_storage
    @deco_token_limit
    def generate(
        self,
        *,
        messages: list[ChatCompletionMessageParam],
        use_backend_cache=True,
        auto_continue=True,
        raw_completion=False,
        **kwargs,
    ):
        logger.debug(f"message:{messages}")
        if "tools" in kwargs:
            tool_funcs = kwargs.get("tools")
            kwargs.update({"tools": batch_make_tools(tool_funcs)})
        try:
            completion = self.client.chat.completions.create(
                messages=messages,
                model=self.ai_model.model,
                temperature=self.ai_model.temperature,
                stream=False,
                **kwargs,
            )
            logger.debug(completion)
            if raw_completion:
                return completion
            result = completion.choices[0].message.content

            match completion.choices[0].finish_reason:
                case "stop":
                    return result
                case "length":
                    if not auto_continue:
                        return result
                    _msgs = [
                        *messages,
                        {"role": "assistant", "content": result},
                    ]
                    result += self.generate(
                        messages=_msgs,
                        use_backend_cache=use_backend_cache,
                        auto_continue=auto_continue,
                        **kwargs,
                    )
                    return result
                case "tool_calls":
                    tool_call_id = completion.choices[0].message.tool_calls[0].id
                    tool_name = (
                        completion.choices[0].message.tool_calls[0].function.name
                    )
                    tool_args = json.loads(
                        completion.choices[0].message.tool_calls[0].function.arguments
                    )
                    tool_message = call_tool(
                        tool_call_id, tool_name, func_list=tool_funcs, **tool_args
                    )
                    logger.debug(
                        f"call function {tool_name}, result message: {tool_message}"
                    )
                    messages.append(
                        completion.choices[0].message.model_dump(exclude_none=True)
                    )
                    messages.append(tool_message)
                    result = self.generate(
                        messages=messages,
                        use_backend_cache=use_backend_cache,
                        auto_continue=auto_continue,
                        **kwargs,
                    )
                    return result
                case "function_call":
                    raise DeprecationWarning("function call is not support, use tools instead")
                    # func_call_id = completion.choices[0].message.tool_calls[0].id
                    # func_name = completion.choices[0].message.tool_calls[0].function
                case "content_filter":
                    ...
                case _:
                    pass
        except Exception as e:
            raise ValueError(e)
        return result

    @deco_storage
    @deco_token_limit
    async def agenerate(
        self,
        *,
        messages: list[ChatCompletionMessageParam],
        use_backend_cache=True,
        auto_continue=True,
        raw_completion=False,
        **kwargs,
    ):
        logger.debug(f"message:{messages}")
        async with self._semaphore:
            try:
                completion = await self.aclient.chat.completions.create(
                    messages=messages,
                    model=self.ai_model.model,
                    temperature=self.ai_model.temperature,
                    stream=False,
                    **kwargs,
                )
                logger.debug(completion)
                if raw_completion:
                    return completion
                result = completion.choices[0].message.content
                match completion.choices[0].finish_reason:
                    case "stop":
                        return result
                    case "length":
                        if not auto_continue:
                            return result
                        _msgs = [
                            *messages,
                            {"role": "assistant", "content": result},
                        ]
                        result += await self.agenerate(
                            messages=_msgs,
                            use_backend_cache=use_backend_cache,
                            auto_continue=auto_continue,
                            **kwargs,
                        )
                        return result
                    case "tool_calls":
                        ...
                    case "function_call":
                        ...
                    case "content_filter":
                        ...
                    case _:
                        pass
            except Exception as e:
                raise ValueError(e)
            return result

    @deco_storage
    @deco_token_limit
    def beta_generate(
        self,
        *,
        messages: list[ChatCompletionMessageParam],
        response_format: T,
        extract_property: str | None = None,
        use_backend_cache=True,
        auto_continue=True,
        raw_completion=False,
        **kwargs,
    ):
        logger.debug(f"message:{messages}")
        try:
            completion = self.client.beta.chat.completions.parse(
                messages=messages,
                model=self.ai_model.model,
                temperature=self.ai_model.temperature,
                response_format=response_format,
                **kwargs,
            )
            if raw_completion:
                return completion
            logger.debug(completion)
            parsed_model: T = completion.choices[0].message.parsed
            if extract_property is None:
                return parsed_model
            result: str = getattr(parsed_model, extract_property)
            match completion.choices[0].finish_reason:
                case "stop":
                    return result
                case "length":
                    if not auto_continue:
                        return result
                    _msgs = [
                        *messages,
                        {"role": "assistant", "content": result},
                    ]
                    result += self.beta_generate(
                        messages=_msgs,
                        response_format=response_format,
                        extract_property=extract_property,
                        use_backend_cache=use_backend_cache,
                        auto_continue=auto_continue,
                        **kwargs,
                    )
                    return result
                case "tool_calls":
                    ...
                case "function_call":
                    func_call_id = completion.choices[0].message.tool_calls[0].id
                    func_name = completion.choices[0].message.tool_calls[0].function
                case "content_filter":
                    ...
                case _:
                    pass
        except Exception as e:
            raise ValueError(e)
        return result

    @deco_storage
    @deco_token_limit
    async def beta_agenerate(
        self,
        *,
        messages: list[ChatCompletionMessageParam],
        response_format: T,
        extract_property: str | None = None,
        use_backend_cache=True,
        auto_continue=True,
        raw_completion=False,
        **kwargs,
    ):
        logger.debug(f"message:{messages}")
        async with self._semaphore:
            try:
                completion = await self.aclient.chat.completions.parse(
                    messages=messages,
                    model=self.ai_model.model,
                    temperature=self.ai_model.temperature,
                    response_format=response_format,
                    **kwargs,
                )
                if raw_completion:
                    return completion
                logger.debug(completion)
                parsed_model: T = completion.choices[0].message.parsed
                if extract_property is None:
                    return parsed_model
                result: str = getattr(parsed_model, extract_property)
                match completion.choices[0].finish_reason:
                    case "stop":
                        return result
                    case "length":
                        if not auto_continue:
                            return result
                        _msgs = [
                            *messages,
                            {"role": "assistant", "content": result},
                        ]
                        result += await self.agenerate(
                            messages=_msgs,
                            response_format=response_format,
                            extract_property=extract_property,
                            use_backend_cache=use_backend_cache,
                            auto_continue=auto_continue,
                            **kwargs,
                        )
                        return result
                    case "tool_calls":
                        ...
                    case "function_call":
                        ...
                    case "content_filter":
                        ...
                    case _:
                        pass
            except Exception as e:
                raise ValueError(e)
            return result



from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        env_file=".env", env_ignore_empty=True, extra="ignore"
    )
    # Redis相关
    REDIS_ENDPOINT: str = "xxxx:xxx"
    REDIS_PASSWORD: str = "xxxx"
    REDIS_URL_PREFIX: str = "rediss"

    # Mongo相关
    MONGO_URL: str = "mongodb+srv://xxxx/"

    MONGO_DB_NAME: str = "xxx"
    MONGO_COLLECTION_NAME: str = "test"

    # 测试模式
    DEBUG: bool = False

    @property
    def REDIS_URL(self):
        return f"{self.REDIS_URL_PREFIX}://default:{self.REDIS_PASSWORD}@{self.REDIS_ENDPOINT}"


settings = Settings()



import jinja2
import re
from .agent import AIAgent
from functools import wraps
from copy import deepcopy
from typing import Literal, Callable
import pathlib
import inspect

_root = pathlib.Path(__file__).parent.parent
_sys_ptn = re.compile("^System:([^|]*)(\|\|)?", flags=re.RegexFlag.IGNORECASE)
_user_ptn = re.compile("User:([^|]*)(\|\|)?", flags=re.RegexFlag.IGNORECASE)
_assistant_ptn = re.compile("Assistant:([^|]*)(\|\|)?", flags=re.RegexFlag.IGNORECASE)
_env = jinja2.Environment(
    loader=jinja2.FileSystemLoader(f"{_root}/templates"),
    autoescape=jinja2.select_autoescape(enabled_extensions=("jinja2",)),
)


def build_message(role: Literal["system", "user", "assitant"], content):
    return {"role": role, "content": content}


def _get_content_text(content_dict: dict):
    _content_dict = deepcopy(content_dict)
    if isinstance(content_dict["content"], re.Match):
        _content_dict["content"] = content_dict["content"].group(1).strip("\n ")
    return _content_dict

def _assign_function_params(func, *args, **kwargs):
    signature = inspect.signature(func)
    parameter_names = signature.parameters.keys()
    result = kwargs.copy()
    for name,arg in zip(parameter_names,args):
        if name not in result:
           result[name] = arg 
    return result

def ai_template(func: Callable | None = None, post_process: Callable | None = None):
    def outer_wrapper(wrap_func: Callable):
        @wraps(wrap_func)
        def wrapper(
            *args,
            _ae_agent: AIAgent | None = None,
            _ae_params: dict | None = None,
            **kwargs,
        ) -> list[dict] | str:
            params = _assign_function_params(wrap_func,*args,**kwargs)
            doc_str = inspect.getdoc(wrap_func)
            if doc_str is None:
                raw_prompt = _env.get_template(f"{wrap_func.__name__}.jinja2").render(
                    **params
                )
            else:
                raw_prompt = jinja2.Template(doc_str).render(**params)
            match_messages: list[dict] = []
            sys_content = _sys_ptn.search(raw_prompt)
            user_content = _user_ptn.search(raw_prompt)
            assistant_content = _assistant_ptn.search(raw_prompt)

            if sys_content:
                match_messages.append(build_message("system", sys_content))
            if user_content:
                match_messages.append(build_message("user", user_content))
            if assistant_content:
                match_messages.append(build_message("assitant", assistant_content))

            match_messages.sort(key=lambda x: x["content"].span()[0])
            messages = list(map(_get_content_text, match_messages))
            if post_process:
                messages = post_process(messages)
            if _ae_agent:
                _ae_params = _ae_params or {}
                result = _ae_agent.generate(messages=messages, **_ae_params)
                return result
            return messages

        return wrapper

    if func is None:
        return outer_wrapper
    return outer_wrapper(func)



from typing import Callable, Literal

def build_message(role: Literal['system', 'user', 'assitant'], content): ...
def ai_template(func: Callable | None = None, post_process: Callable | None = None): ...



from pydantic import BaseModel, HttpUrl, Field
from typing import Literal

__all__ = ["DeepSeek", "GPT", "QWen", "Moonshot"]


class AIModelBase(BaseModel):
    model: str
    base_url: HttpUrl | None = None
    api_key: str | None = Field(
        pattern="^sk-", description="should start with sk", default=None
    )
    temperature: float = Field(
        ge=0.0, le=2.0, description="float between 0 to 2", default=1.0
    )
    tokens_limit: int = 8192
    token_counter: str = "cl100k_base"


class DeepSeek(AIModelBase):
    model: Literal["deepseek-chat", "deepseek-coder"] = "deepseek-chat"
    base_url: HttpUrl | None = "https://api.deepseek.com"
    tokens_limit: int = 32000


class QWen(AIModelBase):
    model: Literal["qwen-long", "qwen-turbo", "qwen-plus", "qwen-max"] = "qwen-long"
    base_url: HttpUrl | None = "https://dashscope.aliyuncs.com/compatible-mode/v1"


class GPT(AIModelBase):
    model: Literal["gpt-4o", "gpt-3.5-turbo"] = "gpt-4o"


class Moonshot(AIModelBase):
    model: Literal["moonshot-v1-8k", "moonshot-v1-32k", "moonshot-v1-128k"] = (
        "moonshot-v1-8k"
    )
    base_url: HttpUrl | None = "https://api.moonshot.cn/v1"



from pydantic import BaseModel
from .agent import AIAgent
import redis


class StorageRedis(BaseModel):
    url: str


def setup_storage(agent: AIAgent, storage_backend: StorageRedis):
    _redis = redis.from_url(storage_backend.url)
    agent._store_backend = _redis



import inspect
from pydantic import create_model
import openai
from typing import Callable


def craft_tool_from_function(func):
    """convert python function to openai tool"""
    signature = inspect.signature(func)

    fields = {}
    for name, param in signature.parameters.items():
        _annotation = str if param.annotation is param.empty else param.annotation
        _default = ... if param.default is param.empty else param.default
        fields[name] = (_annotation,_default)


    docstring = func.__doc__

    model = create_model(func.__name__, __doc__=docstring, **fields)
    tool = openai.pydantic_function_tool(model)

    # 由于openai转换有bug，因此重新赋值required
    tool['function']['parameters']['required'] = model.model_json_schema()['required']
    return tool


def batch_make_tools(funcs) -> list[dict]:
    tools = []
    for func in funcs:
        if isinstance(func, dict):
            tools.append(func)
        elif isinstance(func, Callable):
            tools.append(craft_tool_from_function(func))
        else:
            raise ValueError(f"Wrong type of parameters: {func.__name__}:{type(func)}")
    return tools


def call_tool(tool_call_id, func_name, func_list: list, **kwargs):
    for func in func_list:
        if func.__name__ == func_name:
            return {
                "tool_call_id": tool_call_id,
                "role": "tool",
                "content": str(func(**kwargs)),
            }



import tests
import pytest
from agenteasy.agent import AIAgent
from agenteasy.models import *
from agenteasy.storage import setup_storage, StorageRedis

agent = AIAgent(GPT(model="gpt-3.5-turbo"))
aagent = AIAgent(GPT(model="gpt-3.5-turbo"), model_async=True)


def test_generate():
    content = agent.generate(
        messages=[{"role": "user", "content": "直接返回两个字：通过"}]
    )
    assert "通过" in content


def test_generate_longer():
    content = agent.generate(
        messages=[{"role": "user", "content": "你是谁？"}], max_tokens=8
    )
    assert len(content) > 10


@pytest.mark.asyncio
async def test_agenerate():
    content = await aagent.agenerate(
        messages=[{"role": "user", "content": "直接返回两个字：通过"}]
    )
    assert "通过" in content


def test_backend():
    from agenteasy.config import settings

    bk_agent = AIAgent(GPT(model="gpt-3.5-turbo"))
    setup_storage(bk_agent, StorageRedis(url=settings.REDIS_URL))
    num = bk_agent.generate(
        messages=[{"role": "user", "content": "给个30000以内的随机整数"}],
        use_backend_cache=False,
    )
    num_compare = bk_agent.generate(
        messages=[{"role": "user", "content": "给个30000以内的随机整数"}],
        use_backend_cache=False,
    )
    assert num != num_compare
    num_cache = bk_agent.generate(
        messages=[{"role": "user", "content": "给个30000以内的随机整数"}]
    )
    assert num_cache == num_compare



import logging

# from .config import settings

logger_name = "agentlog"
# FORMAT = "%(levelname)-5s:  [%(asctime)s] %(name)s (%(filename)s - %(funcName)s - %(lineno)d) - %(message)s"
# level = logging.DEBUG if settings.DEBUG else logging.WARNING
# if level == logging.DEBUG:
#     logging.basicConfig(
#         format=FORMAT,
#         encoding="utf-8",
#         datefmt="%Y-%m-%d %H:%M:%S",
#         level=level,
#     )
logger = logging.getLogger(logger_name)



import tests
import pytest
from pydantic import ValidationError
from agenteasy.models import *


def test_api_key_format():
    with pytest.raises(ValidationError):
        Moonshot(api_key="abcd123")
